/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * This file is part of the libetonyek project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef KEY6OBJECTTYPE_H_INCLUDED
#define KEY6OBJECTTYPE_H_INCLUDED

namespace libetonyek
{

namespace KEY6ObjectType
{

enum
{
  Document = 1,
  Presentation = 2,
  SlideList = 4,
  Slide = 5,
  Placeholder = 7,
  SlideStyle = 9,
  Notes = 15
};

}

}

#endif

/* vim:set shiftwidth=2 softtabstop=2 expandtab: */
